/*******************************************************************************
 * Copyright (c) 2012 AGETO Service GmbH and others.
 * All rights reserved.
 *  
 * This program and the accompanying materials are made available under the 
 * terms of the Eclipse Distribution License v1.0 which accompanies this distribution,
 * and is available at https://www.eclipse.org/org/documents/edl-v10.html.
 *
 * Contributors:
 *     Gunnar Wagenknecht - initial API and implementation
 *******************************************************************************/
/*******************************************************************************
 * This is sample code. Use at your own risk. It comes WITHOUT any warranty.
 * Released to public domain. Please copy & paste as you wish.
 *
 * Initial Contribution: Gunnar Wagenknecht
 *******************************************************************************/
package sample.jpa;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.TypedQuery;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.time.DateFormatUtils;
import org.apache.commons.lang.time.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import sample.service.GreetingService;
import sample_imp.service.GreetingServiceImpl;

/**
 * Extension of {@link GreetingServiceImpl} which stores greetings into MongoDB
 * using EclipseLink JPA.
 */
public class JpaGreetingServiceImpl implements GreetingService {

	private final EntityManagerFactory entityManagerFactory;
	private final String myNodeId;
	
	private static final Logger LOG = LoggerFactory.getLogger(GreetingServiceImpl.class);

	/**
	 * Creates a new instance.
	 * 
	 * @param myNodeId
	 * @param entityManagerFactory
	 */
	public JpaGreetingServiceImpl(final String myNodeId, final EntityManagerFactory entityManagerFactory) {
		this.myNodeId = myNodeId;
		this.entityManagerFactory = entityManagerFactory;
	}
	
	public static class Greeting {

		private final long created;
		private final String text;
		private final String submittedBy;
		private long processedOn;
		private String processedBy;

		public Greeting(final long created, final String text, final String submittedBy) {
			this.created = created;
			this.text = text;
			this.submittedBy = submittedBy;
		}

		public long getCreated() {
			return created;
		}

		private String getFormattedDate(final long timestamp) {
			if (DateUtils.isSameDay(new Date(), new Date(timestamp))) {
				return DateFormatUtils.ISO_TIME_NO_T_TIME_ZONE_FORMAT.format(timestamp);
			}
			return DateFormatUtils.SMTP_DATETIME_FORMAT.format(timestamp);
		}

		public String getProcessedBy() {
			return processedBy;
		}

		public long getProcessedOn() {
			return processedOn;
		}

		public String getSubmittedBy() {
			return submittedBy;
		}

		public String getText() {
			return text;
		}

		public void process(final long processedOn, final String processedBy) {
			this.processedOn = processedOn;
			this.processedBy = processedBy;
		}

		@Override
		public String toString() {
			final StringBuilder greeting = new StringBuilder();
			greeting.append(StringUtils.trimToEmpty(text));
			greeting.append(" (");
			greeting.append("submitted by ").append(submittedBy);
			if (processedOn > 0) {
				greeting.append(", processed on ").append(getFormattedDate(processedOn));
				greeting.append(", by ").append(processedBy);
			}
			greeting.append(")");

			return greeting.toString();
		}

	}
	
	@Override
	public Collection<String> getGreetings() throws Exception {
		// load all (sorted based on creation time)
		final List<Greeting> allGreetings = loadAll();

		// generate greetings (but start with latest)
		final List<String> greetings = new ArrayList<String>(allGreetings.size());
		for (int i = allGreetings.size() - 1; i >= 0; i--) {
			greetings.add(allGreetings.get(i).toString());
		}
		return greetings;
	}
	
	@Override
	public void processGreetings() throws Exception {
		if (LOG.isTraceEnabled()) {
			LOG.trace("Processing greetings...");
		}

		// each call processed at most 3 greetings
		// (to have a better demo)
		for (int i = 0; i < 3; i++) {
			final Greeting greeting = nextUnprocessed();
			if (null == greeting) {
				break;
			}
			LOG.info("Processing greeting: {}", greeting);
			greeting.process(System.currentTimeMillis(), myNodeId);
			addProcessed(greeting);
		}

		if (LOG.isTraceEnabled()) {
			LOG.trace("Finished processing greetings.");
		}
	}

	@Override
	public void sayHello(final String greeting) throws Exception {
		if (null == greeting) {
			throw new IllegalArgumentException("greeting must not be null");
		}

		addUnprocessed(new Greeting(System.currentTimeMillis(), greeting, myNodeId));
	}

	
	protected void addProcessed(final Greeting greeting) {
		final EntityManager em = getEntityManagerFactory().createEntityManager();
		final GreetingEntity entity = new GreetingEntity();
		entity.setCreated(greeting.getCreated());
		entity.setSubmittedBy(greeting.getSubmittedBy());
		entity.setText(greeting.getText());
		entity.setProcessedOn(greeting.getProcessedOn());
		entity.setProcessedBy(greeting.getProcessedBy());
		em.getTransaction().begin();
		em.persist(entity);
		em.getTransaction().commit();
		em.close();
	}

	protected void addUnprocessed(final Greeting greeting) {
		final EntityManager em = getEntityManagerFactory().createEntityManager();
		final GreetingEntity entity = new GreetingEntity();
		entity.setCreated(greeting.getCreated());
		entity.setSubmittedBy(greeting.getSubmittedBy());
		entity.setText(greeting.getText());
		em.getTransaction().begin();
		em.persist(entity);
		em.getTransaction().commit();
		em.close();
	}

	public void close() {
		getEntityManagerFactory().close();
	}

	/**
	 * Returns the entityManagerFactory.
	 * 
	 * @return the entityManagerFactory
	 */
	public EntityManagerFactory getEntityManagerFactory() {
		return entityManagerFactory;
	}

	protected List<Greeting> loadAll() {
		final EntityManager em = getEntityManagerFactory().createEntityManager();
		final TypedQuery<GreetingEntity> q = em.createQuery("SELECT g FROM GreetingEntity g ORDER BY g.created desc", GreetingEntity.class);
		final List<GreetingEntity> results = q.getResultList();
		final List<Greeting> greetings = new ArrayList<JpaGreetingServiceImpl.Greeting>(results.size());
		for (final GreetingEntity entity : results) {
			final Greeting greeting = new Greeting(entity.getCreated(), entity.getText(), entity.getSubmittedBy());
			if (entity.getProcessedOn() > 0) {
				greeting.process(entity.getProcessedOn(), entity.getProcessedBy());
			}
			greetings.add(greeting);
		}
		em.close();
		return greetings;
	}

	protected Greeting nextUnprocessed() {
		final EntityManager em = getEntityManagerFactory().createEntityManager();
		final TypedQuery<GreetingEntity> q = em.createQuery("SELECT g FROM GreetingEntity g WHERE g.processedOn = 0 ORDER BY g.created desc", GreetingEntity.class);
		q.setMaxResults(1);
		final List<GreetingEntity> results = q.getResultList();
		if (results.isEmpty()) {
			return null;
		}

		final GreetingEntity entity = results.get(0);
		em.getTransaction().begin();
		final Greeting greeting = new Greeting(entity.getCreated(), entity.getText(), entity.getSubmittedBy());
		if (entity.getProcessedOn() > 0) {
			greeting.process(entity.getProcessedOn(), entity.getProcessedBy());
		}
		em.remove(entity);
		em.getTransaction().commit();
		em.close();
		return greeting;
	}

}
